"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.legacy.saving.serialization import (
    deserialize_keras_object as deserialize_keras_object,
)
from keras.src.legacy.saving.serialization import (
    serialize_keras_object as serialize_keras_object,
)
